/*
 * fixdisktable.c --
 *
 *	This program is indented to fix a corrupted msdos partition
 *	table.	It searches for known filesystem types on the disk,
 *	and then it attempts to regenerate a partition table based
 *	on the table information.  Currently, it looks for EXT2,
 *	FAT16/FAT32, UFS, and NTFS partitions.
 * 
 * Notes:
 *	Bits and pieces of this code, mainly structure definitions,
 *	have been taken from other code such as sfdisk, the Linux
 *	kernel, etc.
 *
 * Thanks to:
 *	Andries E. Brouwer (aeb@cwi.nl) for partition table handling code
 *	Martin von Lwis for NTFS related info
 *	Theodore Ts'o for llseek code
 *
 * Author:
 *	Gordon Chaffee
 *
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/ioctl.h>
#include "fat_io.h"

int heads, cylinders, sectors, cylindersize;
int debug = 0;
int verbose = 0;

typedef signed   char fd_s8;
typedef unsigned char fd_u8;
typedef signed   short fd_s16;
typedef unsigned short fd_u16;
typedef signed   long  fd_s32;
typedef unsigned long  fd_u32;
typedef signed   long long fd_s64;
typedef unsigned long long fd_u64;

typedef struct {
    fd_u8 h;
    fd_u8 s;
    fd_u8 c;
} chs;
chs zero_chs = { 0,0,0 };

typedef struct {
    fd_u32 h;
    fd_u32 s;
    fd_u32 c;
} longchs;
longchs zero_longchs = {0, 0, 0};

/* MSDOS partition table */
struct partition {
    fd_u8 boot_ind;		/* 0x80 - active */
    fd_u8 head;			/* starting head */
    fd_u8 sector;		/* starting sector */
    fd_u8 cyl;			/* starting cylinder */
    fd_u8 sys_ind;		/* What partition type */
    fd_u8 end_head;		/* end head */
    fd_u8 end_sector;		/* end sector */
    fd_u8 end_cyl;		/* end cylinder */
    fd_u32 start_sect;		/* starting sector counting from 0 */
    fd_u32 nr_sects;		/* nr of sectors in partition */
};

/* Generic partition information */
struct partition_info {
    int type;
    fd_s64 offset;
    fd_u64 length;
};

/* Ioctls to read the disk geometry */
#define HDIO_GETGEO 0x0301		/* Linux */
#define DIOCGDINFO  0x41146465		/* FreeBSD */

struct hd_geometry {
      fd_u8 heads;
      fd_u8 sectors;
      fd_u16 cylinders;
      fd_u32 start;
};

/* FreeBSD */
#define BSD_PARTITION		0xa5	/* Partition ID */

#define BSD_DISKMAGIC	(0x82564557UL)	/* The disk magic number */
#define BSD_MAXPARTITIONS	8
#define BSD_FS_UNUSED		0	/* disklabel unused partition entry ID */
struct bsd_disklabel {
	fd_u32	d_magic;		/* the magic number */
	fd_s16	d_type;			/* drive type */
	fd_s16	d_subtype;		/* controller/d_type specific */
	char	d_typename[16];		/* type name, e.g. "eagle" */
	char	d_packname[16];			/* pack identifier */ 
	fd_u32	d_secsize;		/* # of bytes per sector */
	fd_u32	d_nsectors;		/* # of data sectors per track */
	fd_u32	d_ntracks;		/* # of tracks per cylinder */
	fd_u32	d_ncylinders;		/* # of data cylinders per unit */
	fd_u32	d_secpercyl;		/* # of data sectors per cylinder */
	fd_u32	d_secperunit;		/* # of data sectors per unit */
	fd_u16	d_sparespertrack;	/* # of spare sectors per track */
	fd_u16	d_sparespercyl;		/* # of spare sectors per cylinder */
	fd_u32	d_acylinders;		/* # of alt. cylinders per unit */
	fd_u16	d_rpm;			/* rotational speed */
	fd_u16	d_interleave;		/* hardware sector interleave */
	fd_u16	d_trackskew;		/* sector 0 skew, per track */
	fd_u16	d_cylskew;		/* sector 0 skew, per cylinder */
	fd_u32	d_headswitch;		/* head switch time, usec */
	fd_u32	d_trkseek;		/* track-to-track seek, usec */
	fd_u32	d_flags;		/* generic flags */
#define NDDATA 5
	fd_u32	d_drivedata[NDDATA];	/* drive-type specific information */
#define NSPARE 5
	fd_u32	d_spare[NSPARE];	/* reserved for future use */
	fd_u32	d_magic2;		/* the magic number (again) */
	fd_u16	d_checksum;		/* xor of data incl. partitions */

			/* filesystem and partition information: */
	fd_u16	d_npartitions;		/* number of partitions in following */
	fd_u32	d_bbsize;		/* size of boot area at sn0, bytes */
	fd_u32	d_sbsize;		/* max size of fs superblock, bytes */
	struct	bsd_partition {		/* the partition table */
		fd_u32	p_size;		/* number of sectors in partition */
		fd_u32	p_offset;	/* starting sector */
		fd_u32	p_fsize;	/* filesystem basic fragment size */
		fd_u8	p_fstype;	/* filesystem type, see below */
		fd_u8	p_frag;		/* filesystem fragments per block */
		fd_u16	p_cpg;		/* filesystem cylinders per group */
	} d_partitions[BSD_MAXPARTITIONS];	/* actually may be more */
};

#define MAXPARTITIONS 8
struct freebsd_disklabel {
	fd_u32 d_magic;		/* the magic number */
	fd_u16 d_type;		/* drive type */
	fd_u16 d_subtype;		/* controller/d_type specific */
	char	  d_typename[16];	/* type name, e.g. "eagle" */

	/* 
	 * d_packname contains the pack identifier and is returned when
	 * the disklabel is read off the disk or in-core copy.
	 * d_boot0 and d_boot1 are the (optional) names of the
	 * primary (block 0) and secondary (block 1-15) bootstraps
	 * as found in /usr/mdec.  These are returned when using
	 * getdiskbyname(3) to retrieve the values from /etc/disktab.
	 */
	char	  d_packname[16];		/* pack identifier */ 

			/* disk geometry: */
	fd_u32 d_secsize;		/* # of bytes per sector */
	fd_u32 d_nsectors;		/* # of data sectors per track */
	fd_u32 d_ntracks;		/* # of tracks per cylinder */
	fd_u32 d_ncylinders;		/* # of data cylinders per unit */
	fd_u32 d_secpercyl;		/* # of data sectors per cylinder */
	fd_u32 d_secperunit;		/* # of data sectors per unit */

	/*
	 * Spares (bad sector replacements) below are not counted in
	 * d_nsectors or d_secpercyl.  Spare sectors are assumed to
	 * be physical sectors which occupy space at the end of each
	 * track and/or cylinder.
	 */
	fd_u16 d_sparespertrack;	/* # of spare sectors per track */
	fd_u16 d_sparespercyl;	/* # of spare sectors per cylinder */
	/*
	 * Alternate cylinders include maintenance, replacement, configuration
	 * description areas, etc.
	 */
	fd_u32 d_acylinders;		/* # of alt. cylinders per unit */

			/* hardware characteristics: */
	/*
	 * d_interleave, d_trackskew and d_cylskew describe perturbations
	 * in the media format used to compensate for a slow controller.
	 * Interleave is physical sector interleave, set up by the
	 * formatter or controller when formatting.  When interleaving is
	 * in use, logically adjacent sectors are not physically
	 * contiguous, but instead are separated by some number of
	 * sectors.  It is specified as the ratio of physical sectors
	 * traversed per logical sector.  Thus an interleave of 1:1
	 * implies contiguous layout, while 2:1 implies that logical
	 * sector 0 is separated by one sector from logical sector 1.
	 * d_trackskew is the offset of sector 0 on track N relative to
	 * sector 0 on track N-1 on the same cylinder.  Finally, d_cylskew
	 * is the offset of sector 0 on cylinder N relative to sector 0
	 * on cylinder N-1.
	 */
	fd_u16 d_rpm;		/* rotational speed */
	fd_u16 d_interleave;		/* hardware sector interleave */
	fd_u16 d_trackskew;		/* sector 0 skew, per track */
	fd_u16 d_cylskew;		/* sector 0 skew, per cylinder */
	fd_u32 d_headswitch;		/* head switch time, usec */
	fd_u32 d_trkseek;		/* track-to-track seek, usec */
	fd_u32 d_flags;		/* generic flags */
#define NDDATA 5
	fd_u32 d_drivedata[NDDATA];	/* drive-type specific information */
#define NSPARE 5
	fd_u32 d_spare[NSPARE];	/* reserved for future use */
	fd_u32 d_magic2;		/* the magic number (again) */
	fd_u16 d_checksum;		/* xor of data incl. partitions */

			/* filesystem and partition information: */
	fd_u16 d_npartitions;	/* number of partitions in following */
	fd_u32 d_bbsize;		/* size of boot area at sn0, bytes */
	fd_u32 d_sbsize;		/* max size of fs superblock, bytes */
	struct	fbsd_partition {	/* the partition table */
		fd_u32 p_size;	/* number of sectors in partition */
		fd_u32 p_offset;	/* starting sector */
		fd_u32 p_fsize;	/* filesystem basic fragment size */
		fd_u8 p_fstype;	/* filesystem type, see below */
		fd_u8 p_frag;	/* filesystem fragments per block */
		union {
			fd_u16 cpg;	/* UFS: FS cylinders per group */
			fd_u16 sgs;	/* LFS: FS segment shift */
		} fd_partition_u1;
#define	p_cpg	fd_partition_u1.cpg
#define	p_sgs	fd_partition_u1.sgs
	} d_partitions[MAXPARTITIONS];	/* actually may be more */
};


#define PARTITION_TYPE_NONE		0
#define PARTITION_TYPE_FAT16		1
#define PARTITION_TYPE_FAT32		2
#define PARTITION_TYPE_NTFS		3
#define PARTITION_TYPE_EXT2		4
#define PARTITION_TYPE_UFS		5
#define PARTITION_TYPE_BSD_LABEL	6
#define PARTITION_TYPE_LINUXSWAP	7

#define le16_to_cpu(x) x
#define le32_to_cpu(x) x
#define CF_LE_W(x) x
#define CF_LE_L(x) x

#define swap32(x) ((((x) & 0xff000000) >> 24) | \
		   (((x) & 0x00ff0000) >> 8)  | \
		   (((x) & 0x0000ff00) << 8)  | \
		   (((x) & 0x000000ff) << 24))
#define swap16(x) ((((x) & 0x00ff0000) >> 8)  | \
		   (((x) & 0x0000ff00) << 8))

#define GETU64(p) \
    ((((long long) (p)[0]) <<  0) | (((long long) (p)[1]) <<  8) | \
     (((long long) (p)[2]) << 16) | (((long long) (p)[3]) << 24) | \
     (((long long) (p)[4]) << 32) | (((long long) (p)[5]) << 40) | \
     (((long long) (p)[6]) << 48) | (((long long) (p)[7]) << 56))

struct fat_boot_sector {
	fd_u8	jump_code[3];	/* Boot strap short or near jump */
	fd_u8	system_id[8];	/* Name - can be used to special case
				   partition manager volumes */
	fd_u8	sector_size[2];	/* bytes per logical sector */
	fd_u8	cluster_size;	/* sectors/cluster */
	fd_u16	reserved;	/* reserved sectors */
	fd_u8	fats;		/* number of FATs */
	fd_u8	dir_entries[2];	/* root directory entries */
	fd_u8	sectors[2];	/* number of sectors */
	fd_u8	media;		/* media code (unused) */
	fd_u16	fat_length;	/* sectors/FAT */
	fd_u16	secs_track;	/* sectors per track */
	fd_u16	heads;		/* number of heads */
	fd_u32	hidden;		/* hidden sectors (unused) */
	fd_u32	total_sect;	/* number of sectors (if sectors == 0) */

	/* The following fields are only used by FAT32 */
	fd_u32	fat32_length;	/* sectors/FAT */
	fd_u16	flags;		/* bit 8: fat mirroring, low 4: active fat */
	fd_u8	version[2];	/* major, minor filesystem version */
	fd_u32	root_cluster;	/* first cluster in root directory */
	fd_u16	info_sector;	/* filesystem info sector */
	fd_u16	backup_boot;	/* backup boot sector */
	fd_u16	reserved2[6];	/* Unused */
};

struct fat_boot_fsinfo {
	fd_u32   reserved1;	/* Nothing as far as I can tell */
	fd_u32   signature;	/* 0x61417272L */
	fd_u32   free_clusters;	/* Free cluster count.  -1 if unknown */
	fd_u32   next_cluster;	/* Most recently allocated cluster.
				 * Unused under Linux. */
	fd_u32   reserved2[4];
};

#define FD_PAGE_SIZE 4096
union linux_swap_header {
	struct 
	{
		char reserved[FD_PAGE_SIZE - 10];
		char magic[10];
	} magic;
	struct 
	{
		char	     bootbits[1024];	/* Space for disklabel etc. */
		fd_u32 version;
		fd_u32 last_page;
		fd_u32 nr_badpages;
		fd_u32 padding[125];
		fd_u32 badpages[1];
	} info;
};

/*
 * Structure of the super block
 */
struct ext2_super_block {
	fd_u32	s_inodes_count;		/* Inodes count */
	fd_u32	s_blocks_count;		/* Blocks count */
	fd_u32	s_r_blocks_count;	/* Reserved blocks count */
	fd_u32	s_free_blocks_count;	/* Free blocks count */
	fd_u32	s_free_inodes_count;	/* Free inodes count */
	fd_u32	s_first_data_block;	/* First Data Block */
	fd_u32	s_log_block_size;	/* Block size */
	fd_s32	s_log_frag_size;	/* Fragment size */
	fd_u32	s_blocks_per_group;	/* # Blocks per group */
	fd_u32	s_frags_per_group;	/* # Fragments per group */
	fd_u32	s_inodes_per_group;	/* # Inodes per group */
	fd_u32	s_mtime;		/* Mount time */
	fd_u32	s_wtime;		/* Write time */
	fd_u16	s_mnt_count;		/* Mount count */
	fd_s16	s_max_mnt_count;	/* Maximal mount count */
	fd_u16	s_magic;		/* Magic signature */
	fd_u16	s_state;		/* File system state */
	fd_u16	s_errors;		/* Behaviour when detecting errors */
	fd_u16	s_minor_rev_level; 	/* minor revision level */
	fd_u32	s_lastcheck;		/* time of last check */
	fd_u32	s_checkinterval;	/* max. time between checks */
	fd_u32	s_creator_os;		/* OS */
	fd_u32	s_rev_level;		/* Revision level */
	fd_u16	s_def_resuid;		/* Default uid for reserved blocks */
	fd_u16	s_def_resgid;		/* Default gid for reserved blocks */
	/*
	 * These fields are for EXT2_DYNAMIC_REV superblocks only.
	 *
	 * Note: the difference between the compatible feature set and
	 * the incompatible feature set is that if there is a bit set
	 * in the incompatible feature set that the kernel doesn't
	 * know about, it should refuse to mount the filesystem.
	 * 
	 * e2fsck's requirements are more strict; if it doesn't know
	 * about a feature in either the compatible or incompatible
	 * feature set, it must abort and not try to meddle with
	 * things it doesn't understand...
	 */
	fd_u32	s_first_ino; 		/* First non-reserved inode */
	fd_u16   s_inode_size; 		/* size of inode structure */
	fd_u16	s_block_group_nr; 	/* block group # of this superblock */
	fd_u32	s_feature_compat; 	/* compatible feature set */
	fd_u32	s_feature_incompat; 	/* incompatible feature set */
	fd_u32	s_feature_ro_compat; 	/* readonly-compatible feature set */
	fd_u8	s_uuid[16];		/* 128-bit uuid for volume */
	char	s_volume_name[16]; 	/* volume name */
	char	s_last_mounted[64]; 	/* directory where last mounted */
	fd_u32	s_algorithm_usage_bitmap; /* For compression */
	/*
	 * Performance hints.  Directory preallocation should only
	 * happen if the EXT2_COMPAT_PREALLOC flag is on.
	 */
	fd_u8	s_prealloc_blocks;	/* Nr of blocks to try to preallocate*/
	fd_u8	s_prealloc_dir_blocks;	/* Nr to preallocate for dirs */
	fd_u16	s_padding1;
	fd_u32	s_reserved[204];	/* Padding to the end of the block */
};

#define UFS_BSIZE	8192
#define UFS_MINBSIZE	4096
#define UFS_FSIZE	1024
#define UFS_MAXFRAG	(UFS_BSIZE / UFS_FSIZE)
#define UFS_NDADDR 12
#define UFS_NINDIR 3

#define UFS_MAXMNTLEN 512
#define UFS_MAXCSBUFS 32

struct ufs_csum {
	fd_u32	cs_ndir;	/* number of directories */
	fd_u32	cs_nbfree;	/* number of free blocks */
	fd_u32	cs_nifree;	/* number of free inodes */
	fd_u32	cs_nffree;	/* number of free frags */
};

typedef struct _ufsquad {
	fd_u32 val[2];
} ufsquad;

/*
 * This is the actual superblock, as it is laid out on the disk.
 */
struct ufs_superblock {
	fd_u32	fs_link;	/* UNUSED */
	fd_u32	fs_rlink;	/* UNUSED */
	fd_u32	fs_sblkno;
	fd_u32	fs_cblkno;
	fd_u32	fs_iblkno;
	fd_u32	fs_dblkno;
	fd_u32	fs_cgoffset;
	fd_u32	fs_cgmask;
	time_t	fs_time;	/* XXX - check type */
	fd_u32	fs_size;
	fd_u32	fs_dsize;
	fd_u32	fs_ncg;
	fd_u32	fs_bsize;
	fd_u32	fs_fsize;
	fd_u32	fs_frag;
	fd_u32	fs_minfree;
	fd_u32	fs_rotdelay;
	fd_u32	fs_rps;
	fd_u32	fs_bmask;
	fd_u32	fs_fmask;
	fd_u32	fs_bshift;
	fd_u32	fs_fshift;
	fd_u32	fs_maxcontig;
	fd_u32	fs_maxbpg;
	fd_u32	fs_fragshift;
	fd_u32	fs_fsbtodb;
	fd_u32	fs_sbsize;
	fd_u32	fs_csmask;
	fd_u32	fs_csshift;
	fd_u32	fs_nindir;
	fd_u32	fs_inopb;
	fd_u32	fs_nspf;
	fd_u32	fs_optim;
	fd_u32	fs_XXX1;
	fd_u32	fs_interleave;
	fd_u32	fs_trackskew;
	fd_u32	fs_id[2];
	fd_u32	fs_csaddr;
	fd_u32	fs_cssize;
	fd_u32	fs_cgsize;
	fd_u32	fs_ntrak;
	fd_u32	fs_nsect;
	fd_u32	fs_spc;
	fd_u32	fs_ncyl;
	fd_u32	fs_cpg;
	fd_u32	fs_ipg;
	fd_u32	fs_fpg;
	struct ufs_csum fs_cstotal;
	fd_u8	fs_fmod;
	fd_u8	fs_clean;
	fd_u8	fs_ronly;
	fd_u8	fs_flags;
	fd_u8	fs_fsmnt[UFS_MAXMNTLEN];
	fd_u32	fs_cgrotor;
	struct ufs_csum * fs_csp[UFS_MAXCSBUFS];
	fd_u32	fs_cpc;
	fd_u16	fs_opostbl[16][8];	/* old rotation block list head */
	fd_s32	fs_sparecon[55];	/* reserved for future constants */
	fd_s32	fs_state;		/* file system state time stamp */
	ufsquad	fs_qbmask;		/* ~usb_bmask - for use with fd_s64 size */
	ufsquad	fs_qfmask;		/* ~usb_fmask - for use with fd_s64 size */
	fd_s32	fs_postblformat;	/* format of positional layout tables */
	fd_s32	fs_nrpos;		/* number of rotational positions */
	fd_s32	fs_postbloff;		/* (fd_s16) rotation block list head */
	fd_s32	fs_rotbloff;		/* (fd_u8) blocks for each rotation */
	fd_s32	fs_magic;		/* magic number */
	fd_u8	fs_space[1];		/* list of blocks for each rotation */

};

struct msdos_dir_entry {
	fd_s8	name[8],ext[3];	/* name and extension */
	fd_u8	attr;		/* attribute bits */
	fd_u8    lcase;		/* Case for base and extension */
	fd_u8	ctime_ms;	/* Creation time, milliseconds */
	fd_u16	ctime;		/* Creation time */
	fd_u16	cdate;		/* Creation date */
	fd_u16	adate;		/* Last access date */
	fd_u16   starthi;	/* High 16 bits of cluster in FAT32 */
	fd_u16	time,date,start;/* time, date and first cluster */
	fd_u32	size;		/* file size (in bytes) */
};

#define SECTOR_SIZE     512 /* sector size (bytes) */
#define MSDOS_SECTOR_BITS	9 /* log2(SECTOR_SIZE) */
#define MSDOS_DPB	(MSDOS_DPS) /* dir entries per block */
#define MSDOS_DPB_BITS	4 /* log2(MSDOS_DPB) */
#define MSDOS_DPS	(SECTOR_SIZE/sizeof(struct msdos_dir_entry))
#define MSDOS_DPS_BITS	4 /* log2(MSDOS_DPS) */
#define MSDOS_DIR_BITS	5 /* log2(sizeof(struct msdos_dir_entry)) */
#define MSDOS_MAX_EXTRA	3
#define ROUND_TO_MULTIPLE(n,m) ((n) && (m) ? (n)+(m)-1-((n)-1)%(m) : 0)

static fd_u64
fat_partition(struct partition_info *pi, unsigned char *buf)
{
    struct fat_boot_sector *bs = (struct fat_boot_sector *) buf;
    int probability = 0;
    int size;
    int data_sectors,logical_sector_size,sector_mult,fat_clusters=0;
    int fat32;

    fd_u16 cluster_size;			/* sectors/cluster */
    fd_u8  fat_bits;			/* FAT bits (12, 16, or 32) */
    fd_u16 fat_length;			/* FAT start & length (sec.) */
    fd_u8  dir_start,dir_entries;	/* root dir start & entries */
    fd_u16 data_start;			/* first data sector */
    fd_u32 clusters;			/* number of clusters */
    fd_u32 fsinfo_offset;		/* FAT32 fsinfo offset */
    fd_u64 length;


    if (bs->media != 0xf8) return 0;
    if (bs->sector_size[0] != 0) return 0;
    size = bs->sector_size[1];
    if (size != 2 && size != 4 && size != 8 && size != 16 &&
	size != 32 && size != 64 && size != 128)
    {
	return 0;
    }
    if ((bs->jump_code[0] == 0xeb) && (bs->jump_code[1] == 0x3c) &&
	(bs->jump_code[2] == 0x90))
    {
	probability++;
    }
    if (strstr("Couldn't find NTLDR", buf) == 0) probability++;
    if (strncmp(bs->system_id, "mkdosfs", 7) == 0) probability++;
    if (strstr("Please insert a bootable floppy", buf) == 0) probability++;

    if (probability < 2) {
	return 0;
    }

    logical_sector_size = bs->sector_size[0] | (bs->sector_size[1] << 8);
    sector_mult = logical_sector_size >> MSDOS_SECTOR_BITS;
    cluster_size = bs->cluster_size*sector_mult;
    if (! bs->fats) return 0;

    if (!bs->fat_length && bs->fat32_length) {
	struct fat_boot_fsinfo *fsinfo;

	/* Must be FAT32 */
	fat32 = 1;
	fat_length= CF_LE_W(bs->fat32_length)*sector_mult;
	fsinfo_offset = CF_LE_W(bs->info_sector) * logical_sector_size + 0x1e0;
	fsinfo = (struct fat_boot_fsinfo *) &buf[fsinfo_offset];
	if (CF_LE_L(fsinfo->signature) != 0x61417272) return 0;
    } else {
	fat32 = 0;
	fat_length = CF_LE_W(bs->fat_length)*sector_mult;
    }
    dir_start= CF_LE_W(bs->reserved)*sector_mult + bs->fats*fat_length;
    dir_entries = bs->dir_entries[0] | (bs->dir_entries[1] << 8);
    data_start = dir_start+ROUND_TO_MULTIPLE((
	dir_entries << MSDOS_DIR_BITS) >> MSDOS_SECTOR_BITS, sector_mult);
    data_sectors = bs->sectors[0] | (bs->sectors[1] << 8);
    if (!data_sectors) {
	data_sectors = CF_LE_L(bs->total_sect);
    }
    data_sectors = data_sectors * sector_mult - data_start;
    if (!bs->cluster_size || !sector_mult) return 0;

    clusters = bs->cluster_size ? data_sectors/
	bs->cluster_size/sector_mult : 0;
    fat_bits = fat32 ? 32 : 16;
    fat_clusters = fat_length * SECTOR_SIZE*8 / fat_bits;

    if (dir_entries & (MSDOS_DPS-1)) return 0;
#if 0
    if (clusters+2 > fat_clusters + MSDOS_MAX_EXTRA) return 0;
#endif
    if (logical_sector_size & (SECTOR_SIZE-1)) return 0;
    if (!bs->secs_track) return 0;
    if (!bs->heads) return 0;

    pi->type = fat32 ? PARTITION_TYPE_FAT32 : PARTITION_TYPE_FAT16;

    length = clusters;
    length *= bs->cluster_size;
    length *= SECTOR_SIZE;

    return length;
}

static int
ntfs_partition(unsigned char *buf)
{
    fd_u32 data;
    fd_u16 blocksize, clusterfactor;
    fd_u32 clustersize;
    fd_u64 size;

    if (strncmp(&buf[3], "NTFS", 4) != 0) return 0;
    data = buf[0x40] | (buf[0x41]<<8) | (buf[0x42]<<16) | (buf[0x43]<<24);
    if (data > 256) return 0;
    data = buf[0x44] | (buf[0x45]<<8) | (buf[0x46]<<16) | (buf[0x47]<<24);
    if (data > 256) return 0;

    blocksize = buf[0x0b] | (buf[0x0c] << 8);
    clusterfactor = buf[0x0d] | (buf[0x0e] << 8);
    clustersize = clusterfactor * blocksize;

    size = GETU64(&buf[0x28]);

    return size * clustersize;
}

#define EXT2_SUPER_MAGIC	0xEF53
#define EXT2_GOOD_OLD_REV	0	/* The good old (original) format */
#define EXT2_DYNAMIC_REV	1 	/* V2 format w/ dynamic inode sizes */
#define EXT2_GOOD_OLD_INODE_SIZE 128
static fd_u64
ext2_partition(unsigned char *buf)
{
    struct ext2_super_block *sb;
    fd_u32 blocks, blocksize_bits, blocksize, first_data_block;
    fd_u64 length;

    sb = (struct ext2_super_block *) buf;
    if (le16_to_cpu(sb->s_magic) != EXT2_SUPER_MAGIC) return 0;

    blocks = le32_to_cpu(sb->s_blocks_count);
    blocksize_bits = le32_to_cpu(sb->s_log_block_size) + 10;
    blocksize = 1 << blocksize_bits;
    if (blocksize != 512 && blocksize != 1024 &&
	blocksize != 2048 && blocksize != 4096)
    {
	return 0;
    }
    first_data_block = le32_to_cpu(sb->s_first_data_block);
    if (le32_to_cpu(sb->s_rev_level) == EXT2_DYNAMIC_REV) {
	if (le16_to_cpu(sb->s_inode_size) != EXT2_GOOD_OLD_INODE_SIZE) {
	    return 0;
	}
    } else if (le32_to_cpu(sb->s_rev_level) != EXT2_GOOD_OLD_REV) {
	return 0;
    }

    length = blocks;
    length *= blocksize;
    return length;
}

static fd_u64
linux_swap_partition(unsigned char *buf)
{
    union linux_swap_header *swap_header = (union linux_swap_header *) buf;
    fd_u64 length;
    
    if (memcmp("SWAP-SPACE",swap_header->magic.magic,10) == 0) {
	/*
	 * This type of swap space doesn't tell us anything about
	 * its length, so we are kind of hose.  We'll need to search
	 * through the entire thing
	 */
	printf("Found Linux swap header\n");
	return sizeof(*swap_header);
    }
    if (memcmp("SWAPSPACE2",swap_header->magic.magic,10) != 0) {
	return 0;
    }
    length = swap_header->info.last_page;
    length *= FD_PAGE_SIZE;
    return length;
}

#define UFS_MAGIC 0x00011954
#define UFS_CIGAM 0x54190100 /* byteswapped MAGIC */
static fd_u64
ufs_partition(unsigned char *buf)
{
    struct ufs_superblock *sb;
    int swap = 0;
    fd_u32 bsize, fsize;
    fd_u32 blocks, blocksize;
    fd_u64 length;
    int i, c;

    sb = (struct ufs_superblock *) buf;
    if (sb->fs_magic != UFS_MAGIC && sb->fs_magic != UFS_CIGAM) return 0;
    if (sb->fs_magic == UFS_CIGAM) swap = 1;
    bsize = swap ? swap32(sb->fs_bsize) : sb->fs_bsize;
    if (bsize != 4096 && bsize != 8192) return 0;
    fsize = swap ? swap32(sb->fs_fsize) : sb->fs_fsize;
    if (fsize != 512 && fsize != 1024) return 0;

    if (sb->fs_fsmnt[0] != '/') return 0;

    if (debug) { printf("Possible UFS partition detected\n"); }
    for (i = 0; i < sb->fs_fsmnt[i]; i++) {
	c = sb->fs_fsmnt[i];
	if (!c) break;
	if (!isprint(c)) return 0;
    }

    printf("Last UFS mount: %s\n", sb->fs_fsmnt);

    blocks = swap ? swap32(sb->fs_dsize) : sb->fs_dsize;
    blocksize = swap ? swap32(sb->fs_size) : sb->fs_size;

    length = blocks;
    length *= blocksize;
    return length;
}

static fd_u64
bsd_disklabel_partition(unsigned char *buf)
{
    struct bsd_disklabel *l;
    struct bsd_partition *p;
    fd_s64 length;

    l = (struct bsd_disklabel *) buf;
    if (l->d_magic != BSD_DISKMAGIC) return 0;
    if (l->d_magic2 != BSD_DISKMAGIC) return 0;

    length = 0;
    p = &l->d_partitions[0];
    while (p - &l->d_partitions[0] <= BSD_MAXPARTITIONS) {
	if (p->p_fstype != BSD_FS_UNUSED) {
	    length = p->p_offset + p->p_size;
	    length *= SECTOR_SIZE;
	}
	p++;
    }
    return length;
}

chs
longchs_to_chs (longchs aa)
{
    chs a;

    if (aa.h < 256 && aa.s < 64 && aa.c < 1024) {
	a.h = aa.h;
	a.s = aa.s | ((aa.c >> 2) & 0xc0);
	a.c = (aa.c & 0xff);
    } else if (heads && sectors) {
	a.h = heads - 1;
	a.s = sectors | 0xc0;
	a.c = 0xff;
    } else
      a = zero_chs;
    return a;
}

longchs
chs_to_longchs (chs a)
{
    longchs aa;

    aa.h = a.h;
    aa.s = (a.s & 0x3f);
    aa.c = (a.s & 0xc0);
    aa.c = (aa.c << 2) + a.c;
    return aa;
}

longchs
ulong_to_longchs (unsigned long sno) {
    longchs aa;

    if (heads && sectors && cylindersize) {
	aa.s = 1 + sno % sectors;
	aa.h = (sno / sectors) % heads;
	aa.c = sno / cylindersize;
	return aa;
    } else {
	return zero_longchs;
    }
}

unsigned long
longchs_to_ulong (longchs aa) {
    return (aa.c*cylindersize + aa.h*sectors + aa.s - 1);
}

chs
ulong_to_chs (unsigned long sno) {
    return longchs_to_chs(ulong_to_longchs(sno));
}

unsigned long
chs_to_ulong (chs a) {
    return longchs_to_ulong(chs_to_longchs(a));
}

#define set_hsc(h,s,c,sector) \
{ \
      fd_u32 tmpsec = sector; \
      s = tmpsec % sectors + 1; \
      tmpsec /= sectors;        \
      h = tmpsec % heads;       \
      tmpsec /= heads;          \
      c = tmpsec & 0xFF;        \
      s |= (tmpsec >> 2) & 0xC0;\
}

void
fill_part_table(struct partition *p, fd_u32 start, fd_u32 end)
{
    int sects;

    p->start_sect = start;
    p->nr_sects = end - start + 1;

    sects = (start / (sectors*heads) > 1023) ? heads*sectors*1024 - 1 : start;
    set_hsc(p->head, p->sector, p->cyl, sects);
    sects = ((end/(sectors*heads) > 1023) ? heads*sectors*1024 - 1 : end);
    set_hsc(p->end_head, p->end_sector, p->end_cyl, sects);
}

static fd_u64
find_partition(struct partition_info *pi, fd_s64 offset, unsigned char *buf)
{
    fd_u64 length, start_sector, end_sector, lsectors;
    longchs start_lchs, end_lchs;
    double flen;
    char *fs = NULL;

    if (debug) { printf("F"); fflush(stdout); }
    length = fat_partition(pi, buf);
    if (debug) { printf("f"); fflush(stdout); }
    if (length) {
	fs = "FAT      ";
	/* pi->type is set in the call to fat_partition */
	offset -= 1 * SECTOR_SIZE;
    }
    if (!length) {
	if (debug) { printf("E"); fflush(stdout); }
	length = ext2_partition(buf);
	if (debug) { printf("e"); fflush(stdout); }
	if (length) {
	    fs = "EXT2     ";
	    pi->type = PARTITION_TYPE_EXT2;
	    offset -= 3 * SECTOR_SIZE;
	}
    }
    if (!length) {
	if (debug) { printf("S"); fflush(stdout); }
	length = linux_swap_partition(buf);
	if (debug) { printf("s"); fflush(stdout); }
	if (length) {
	    fs = "LinuxSwap";
	    pi->type = PARTITION_TYPE_LINUXSWAP;
	}
    }
    if (!length) {
	if (debug) { printf("N"); fflush(stdout); }
	length = ntfs_partition(buf);
	if (debug) { printf("n"); fflush(stdout); }
	if (length) {
	    fs = "NTFS     ";
	    pi->type = PARTITION_TYPE_NTFS;
	    offset -= 1 * SECTOR_SIZE;
	}
    }
    if (!length) {
	if (debug) { printf("B"); fflush(stdout); }
	length = bsd_disklabel_partition(buf);
	if (debug) { printf("b"); fflush(stdout); }
	if (length) {
	    fs = "BSD Label";
	    pi->type = PARTITION_TYPE_BSD_LABEL;
	    offset -= 1 * SECTOR_SIZE;
	}
    }

    if (!length) {
	if (debug) { printf("U"); fflush(stdout); }
	length = ufs_partition(buf);
	if (debug) { printf("u"); fflush(stdout); }
	if (length) {
	    fs = "UFS      ";
	    pi->type = PARTITION_TYPE_UFS;
#if 0
	    offset -= 16 * SECTOR_SIZE;
#endif
	}
    }
    if (length) {

	flen = (double) length;
	flen = flen / ((double) (1 << 20));
	if (offset < 0) {
	    if (verbose) {
		printf("Offset was less than zero\n");
	    }
	    offset = 0;
	}
	if (debug) { printf("\n\n"); }
	printf("%s partition at offset %12.0f, length=(%.03f MB) %.0f\n",
	       fs, (double) offset, flen, (double) length);
	start_sector = offset / SECTOR_SIZE;
	lsectors = length / SECTOR_SIZE;
	end_sector = start_sector + lsectors - 1;
	start_lchs = ulong_to_longchs(start_sector);
	end_lchs = ulong_to_longchs(end_sector);
	printf("   Sectors:    start=%10.0f, end=%10.0f, length=%10.0f\n",
	      (double) start_sector, (double) end_sector, (double) lsectors);
	printf("   Hd,Sec,Cyl: start(%ld,%ld,%ld) end(%ld,%ld,%ld)\n",
	       start_lchs.h, start_lchs.s, start_lchs.c,
	       end_lchs.h, end_lchs.s, end_lchs.c);
	pi->offset = offset;
	pi->length = length;
	if (debug) { printf("\n"); sleep(3); }
	return length;
    }
    return 0;
}

#define EXTRASPACE 8192

int
main(int argc, char **argv)
{
    unsigned char buf[32768];
    int bufpos;
    int fd;
    char *device;
    int n, readsize;
    fd_s64 offset, new_offset, end_offset, new_pos, seek_size;
    fd_u64 length;
    fd_u32 start, end, prev_start;
    int i, j, c;
    struct hd_geometry hdprm;
    struct partition table[50];	// Excessive size since it is static
    struct partition_info tableinfo[50];
    int pos = 0;
    int raw = 0;
    int arg = 1;
    struct freebsd_disklabel disklabel;

    if (argc == 1) goto Usage;

    while (argv[arg] && argv[arg][0] == '-') {
	if (strcmp(argv[arg], "--raw") == 0 ||
	    strcmp(argv[arg], "-r") == 0)
	{
	    raw = 1;
	} else if (strcmp(argv[arg], "--verbose") == 0 ||
	    strcmp(argv[arg], "-v") == 0)
	{
	    verbose = 1;
	} else if (strcmp(argv[arg], "--debug") == 0 ||
	    strcmp(argv[arg], "-d") == 0)
	{
	    debug = 1; verbose = 1;
	} else {
	    goto Usage;
	}
	arg++;
    }

    if (argc == arg) {
    Usage:
	fprintf(stderr, "Usage: fixdisktable ?-d? ?-r? ?-v? <device>\n");
	fprintf(stderr, "          -d (--debug)    Print debugging info\n");
	fprintf(stderr, "          -r (--raw)      Do not skip first track\n");
	fprintf(stderr, "          -v (--verbose)  More verbose\n");
	exit(1);
    }

    device = argv[arg];
    fd = open(device, O_RDONLY);
    if (fd < 0) {
	fprintf(stderr, "Unable to open \'%s\': %s\n",
		device, strerror(errno));
	exit(1);
    }

    if (debug) {
	printf("Getting hard disk geometry\n");
    }
    offset = 0;
    if (ioctl(fd,HDIO_GETGEO,&hdprm) >= 0) {
	heads = hdprm.heads;
	cylinders = hdprm.cylinders;
	sectors = hdprm.sectors;
	printf("cylinders=%d, heads=%d, sectors=%d\n",
	       cylinders, heads, sectors);
    } else if (ioctl(fd, DIOCGDINFO, &disklabel) >= 0) {
	cylinders = disklabel.d_ncylinders;
	heads = disklabel.d_ntracks;
	sectors = disklabel.d_nsectors;
	printf("cylinders=%d, heads=%d, sectors=%d\n",
	       cylinders, heads, sectors);
    } else {
	printf("Unable to get hard disk information.  Working without it.\n");
	heads = cylinders = sectors = 0;
    }
    cylindersize = heads * sectors;

    seek_size = 1;
    seek_size <<= 32;
    end_offset = 0;
    while (seek_size > SECTOR_SIZE) {
	while (1) {
	    new_pos = fat_llseek(fd, end_offset + seek_size, SEEK_SET);
	    if (new_pos > 0) {
		new_pos = fat_llseek(fd, 0, SEEK_CUR);
		if (new_pos > 0 && new_pos > end_offset) {
		    end_offset += seek_size;
		} else {
		    break;
		}
	    } else {
		break;
	    }
	}
	seek_size >>= 1;
    }
    if (verbose) {
	printf("end_offset: %.0f\n", (double) end_offset);
    }
    fat_llseek(fd, 0, SEEK_SET);

    if (! raw) {
	offset = sectors * SECTOR_SIZE;
    }
    if (offset) {
	if (fat_llseek(fd, offset, SEEK_CUR) < 0) {
	    goto done;
	}
    }

    n = read(fd, buf, sizeof(buf));
    while (1) {
        readsize = sizeof(buf) - EXTRASPACE;
	for (i = 0; i < readsize; i += SECTOR_SIZE, offset += SECTOR_SIZE) {
	    length = find_partition(&tableinfo[pos], offset, &buf[i]);
	    if (length) {
		pos++;
		if (length > 1 << 20) {
		    length -= 1 << 20;
		}
		new_offset = length + offset;
		if (new_offset < 0 || new_offset < offset) goto done;
		if (new_offset < length) goto done;

		new_pos = fat_llseek(fd, (fat_loff_t) new_offset, SEEK_SET);
		if (new_pos != new_offset) goto done;
		new_pos = fat_llseek(fd, 0, SEEK_CUR);
		if (new_pos != new_offset) goto done;

		offset = new_offset;
		readsize = sizeof(buf);
		break;
	    }
	}

	bufpos = sizeof(buf) - readsize;
	memcpy(buf, &buf[readsize], bufpos);
	offset = fat_llseek(fd, 0, SEEK_CUR);
	offset -= bufpos;
	n = read(fd, &buf[bufpos], readsize);
	if (n != readsize) {
	    break;
	}
    }

done:
    close(fd);

    if (debug) { printf("\nDone searching for partitions.\n"); }

    memset(table, 0, sizeof(table));
    end = 0;
    start = 0;

    if (pos == 0) {
	printf("No partitions found\n");
	exit(1);
    }

    for (i = 0; i < pos; i++) {
	table[i].boot_ind = 0;

	switch (tableinfo[i].type) {
	case PARTITION_TYPE_FAT16:	table[i].sys_ind = 0x6;  break;
	case PARTITION_TYPE_FAT32:	table[i].sys_ind = 0xb;  break;
	case PARTITION_TYPE_NTFS:	table[i].sys_ind = 0x7;  break;
	case PARTITION_TYPE_EXT2:	table[i].sys_ind = 0x83; break;
	case PARTITION_TYPE_UFS:   	table[i].sys_ind = 0xa5; break;
	case PARTITION_TYPE_BSD_LABEL:	table[i].sys_ind = 0xa5; break;
	case PARTITION_TYPE_LINUXSWAP:	table[i].sys_ind = 0x82; break;
	}

	prev_start = start;
	start = tableinfo[i].offset / SECTOR_SIZE;
	if (i > 0) {
	    /* Remove overlaps and fill small gaps */
	    if (start < end || (start - end < 50000) || table[i-1].sys_ind == 0x82) {
		end = start-1;
		fill_part_table(&table[i-1], prev_start, end);
	    }
	}
	end = start + (tableinfo[i].length / SECTOR_SIZE);

	if (i == 0) {
	    if (start > sectors && sectors + 50 > start) {
		start = sectors;
	    } else if (start < sectors) {
		start = sectors;
	    }
	} else if (i == pos - 1) {
	    fd_s64 end_sector = end_offset / SECTOR_SIZE;
	    if (end_sector - end < 50000 || table[i].sys_ind == 0x82) {
		if (end_sector > end) {
		    end = end_sector;
		}
	    }
	}

	fill_part_table(&table[i], start, end);
    }

    table[0].boot_ind = 0x80;

    if (pos > 4) {
	printf("\nWarning!!:\n"
	       "    There are more than 4 partitions on the disk.\n"
	       "    This program can only recreate primary partitions.\n"
	       "    Write down the numbers that it shows, and you might\n"
	       "    be able to recover using these numbers and fdisk.\n");
    }

    printf("\nNr AF  Hd Sec  Cyl  Hd Sec  Cyl      Start       Size ID\n");
    for (i = 0; i < pos; i++) {
	chs start_chs, end_chs;
	longchs start_lchs, end_lchs;

	start_chs.c = table[i].cyl;
	start_chs.h = table[i].head;
	start_chs.s = table[i].sector;

	end_chs.c = table[i].end_cyl;
	end_chs.h = table[i].end_head;
	end_chs.s = table[i].end_sector;

	start_lchs = chs_to_longchs(start_chs);
	end_lchs = chs_to_longchs(end_chs);

	printf(" %d %2x %3ld %3ld %4ld %3ld %3ld %4ld %10ld %10ld %02x  %s\n",
	       i+1, table[i].boot_ind,
	       start_lchs.h, start_lchs.s, start_lchs.c,
	       end_lchs.h, end_lchs.s, end_lchs.c,
	       table[i].start_sect, table[i].nr_sects, table[i].sys_ind,
	       verbose ? "(Interpretted)" : "");

	if (verbose) {
	    printf(" %d %2x %3d %3d %4d %3d %3d %4d %10ld %10ld %02x  (RAW)\n",
		   i+1, table[i].boot_ind,
		   table[i].head, table[i].sector, table[i].cyl, 
		   table[i].end_head, table[i].end_sector, table[i].end_cyl, 
		   table[i].start_sect, table[i].nr_sects, table[i].sys_ind);
	}

    }

    if (verbose) {
	printf("\n");
	n = sizeof(struct partition) * 4;
	memcpy(buf, table, n);
	for (c = 1, j = 0; j < n; j += sizeof(struct partition), c++) {
	    printf("%d: ", c);
	    for (i = j; i < j + 16; i += 2) {
		printf("%02x%02x ", buf[i], buf[i+1]);
	    }
	    printf("\n");
	}
    }

    printf("\nDo you wish to write this partition table to disk (yes/no)? ");

    fflush(stdout);
    fgets(buf, sizeof(buf), stdin);
    if ('y' == tolower(buf[0]) &&
	'e' == tolower(buf[1]) &&
	's' == tolower(buf[2]))
    {
	printf("\n");

	fd = open(device, O_WRONLY);
	if (lseek(fd, 0x1be, SEEK_SET) < 0) {
	    printf("Unable to seek to partition table location.\n");
	    printf("Exiting without writing partition table.\n");
	    exit(1);
	}

	/* Okay folks, writing is turned on here. */
#if 1
	n = write(fd, table, sizeof(struct partition) * 4);
	if (n != sizeof(struct partition) * 4) {
	    printf("Failed while writing partition table to disk: %s\n",
		   strerror(errno));
	    exit(1);
	}

	buf[0] = 0x55; buf[1] = 0xaa;
	n = write(fd, buf, 2);
	if (n != 2) {
	    printf("Error writing partition table id to disk: %s\n",
		   strerror(errno));
	    exit(1);
	}
#endif
	close(fd);
	printf("Successfully wrote salvaged partition table.\n\n"
	       "You might want to make sure the last partition takes up\n"
	       "the entire last portion of the disk since sometimes it is\n"
	       "recreated a little bit too small.  You can use fdisk to\n"
	       "delete the partition and recreate it.\n\n"
	       "Press <Return> to exit.\n");
	fgets(buf, sizeof(buf), stdin);
    }

    return 0;
}
